import ast
from enum import Enum
from enum import Flag
from functools import singledispatch
from types import FunctionType
from unittest import mock

real_repr = repr


class HasRepr:
    """This class is used for objects where `__repr__()` returns an non-
    parsable representation.

    HasRepr uses the type and repr of the value for equal comparison.

    You can change `__repr__()` to return valid python code or use
    `@customize_repr` to customize repr which is used by inline-
    snapshot.
    """

    def __init__(self, type, str_repr: str) -> None:
        self._type = type
        self._str_repr = str_repr

    def __repr__(self):
        return f"HasRepr({self._type.__qualname__}, {self._str_repr!r})"

    def __eq__(self, other):
        if isinstance(other, HasRepr):
            if other._type is not self._type:
                return False
        else:
            if type(other) is not self._type:
                return False

        other_repr = code_repr(other)
        return other_repr == self._str_repr or other_repr == repr(self)


def used_hasrepr(tree):
    return [
        n
        for n in ast.walk(tree)
        if isinstance(n, ast.Call)
        and isinstance(n.func, ast.Name)
        and n.func.id == "HasRepr"
        and len(n.args) == 2
    ]


@singledispatch
def code_repr_dispatch(value):
    return real_repr(value)


def customize_repr(f):
    """Register a function which should be used to get the code representation
    of a object.

    ``` python
    @customize_repr
    def _(obj: MyCustomClass):
        return f"MyCustomClass(attr={repr(obj.attr)})"
    ```

    it is important to use `repr()` inside the implementation, because it is mocked to return the code representation

    you dont have to provide a custom implementation if:
    * __repr__() of your class returns a valid code representation,
    * and __repr__() uses `repr()` to get the representation of the child objects
    """
    code_repr_dispatch.register(f)


def code_repr(obj):

    with mock.patch("builtins.repr", mocked_code_repr):
        return mocked_code_repr(obj)


def mocked_code_repr(obj):
    from inline_snapshot._adapter.adapter import get_adapter_type

    adapter = get_adapter_type(obj)
    assert adapter is not None
    return adapter.repr(obj)


def value_code_repr(obj):
    if not type(obj) == type(obj):  # pragma: no cover
        # this was caused by https://github.com/samuelcolvin/dirty-equals/issues/104
        # dispatch will not work in cases like this
        return (
            f"HasRepr({repr(type(obj))}, '< type(obj) can not be compared with == >')"
        )

    result = code_repr_dispatch(obj)

    try:
        ast.parse(result)
    except SyntaxError:
        return real_repr(HasRepr(type(obj), result))

    return result


# -8<- [start:Enum]
@customize_repr
def _(value: Enum):
    return f"{type(value).__qualname__}.{value.name}"


# -8<- [end:Enum]


@customize_repr
def _(value: Flag):
    name = type(value).__qualname__
    return " | ".join(f"{name}.{flag.name}" for flag in type(value) if flag in value)


def sort_set_values(set_values):
    is_sorted = False
    try:
        set_values = sorted(set_values)
        is_sorted = True
    except TypeError:
        pass

    set_values = list(map(repr, set_values))
    if not is_sorted:
        set_values = sorted(set_values)

    return set_values


@customize_repr
def _(value: set):
    if len(value) == 0:
        return "set()"

    return "{" + ", ".join(sort_set_values(value)) + "}"


@customize_repr
def _(value: frozenset):
    if len(value) == 0:
        return "frozenset()"

    return "frozenset({" + ", ".join(sort_set_values(value)) + "})"


@customize_repr
def _(value: type):
    return value.__qualname__


@customize_repr
def _(value: FunctionType):
    return value.__qualname__



from __future__ import annotations

import logging
import pathlib
import sys
from collections import defaultdict
from collections.abc import Iterable
from dataclasses import dataclass
from difflib import unified_diff
from itertools import islice

import asttokens.util
from asttokens import LineNumbers

from ._format import enforce_formatting
from ._format import format_code

if sys.version_info >= (3, 10):
    from itertools import pairwise
else:
    from itertools import tee

    def pairwise(iterable):  # type: ignore
        a, b = tee(iterable)
        next(b, None)
        return zip(a, b)


@dataclass(order=True)
class SourcePosition:
    lineno: int
    col_offset: int

    def offset(self, line_numbers):
        return line_numbers.line_to_offset(self.lineno, self.col_offset)


@dataclass(order=True)
class SourceRange:
    start: SourcePosition
    end: SourcePosition

    def __post_init__(self):
        if self.start > self.end:
            raise ValueError("range start should be lower then end")


@dataclass(order=True)
class Replacement:
    range: SourceRange
    text: str
    change_id: int = 0


def start_of(obj) -> SourcePosition:
    if isinstance(obj, asttokens.util.Token):
        return SourcePosition(lineno=obj.start[0], col_offset=obj.start[1])

    if isinstance(obj, SourcePosition):
        return obj

    if isinstance(obj, SourceRange):
        return obj.start

    if isinstance(obj, tuple) and len(obj) == 2:
        return SourcePosition(lineno=obj[0], col_offset=obj[1])

    assert False


def end_of(obj) -> SourcePosition:
    if isinstance(obj, asttokens.util.Token):
        return SourcePosition(lineno=obj.end[0], col_offset=obj.end[1])

    if isinstance(obj, SourceRange):
        return obj.end

    return start_of(obj)


def range_of(obj):
    if isinstance(obj, tuple) and len(obj) == 2:
        return SourceRange(start_of(obj[0]), end_of(obj[1]))

    return SourceRange(start_of(obj), end_of(obj))


class UsageError(Exception):
    pass


class Change:  # ChangeSet
    _next_change_id = 0

    def __init__(self, change_recorder):
        self.change_recorder = change_recorder

        self.change_recorder._changes.append(self)

        self.change_id = self._next_change_id
        type(self)._next_change_id += 1

    def replace(self, node, new_contend, *, filename):
        assert isinstance(new_contend, str)

        self._replace(
            filename,
            range_of(node),
            new_contend,
        )

    def delete(self, node, *, filename):
        self.replace(node, "", filename=filename)

    def insert(self, node, new_content, *, filename):
        self.replace(start_of(node), new_content, filename=filename)

    def _replace(self, filename, range, new_contend):
        source = self.change_recorder.get_source(filename)
        source.replacements.append(
            Replacement(range=range, text=new_contend, change_id=self.change_id)
        )
        source._check()


class SourceFile:
    def __init__(self, filename: pathlib.Path):
        self.replacements: list[Replacement] = []
        self.filename = filename
        self.source = self.filename.read_text("utf-8")

    def rewrite(self):
        new_code = self.new_code()

        with open(self.filename, "bw") as code:
            code.write(new_code.encode())

    def virtual_write(self):
        self.source = self.new_code()

    def _check(self):
        replacements = list(self.replacements)
        replacements.sort()

        for r in replacements:
            assert r.range.start <= r.range.end, r

        for lhs, rhs in pairwise(replacements):
            assert lhs.range.end <= rhs.range.start, (lhs, rhs)

    def new_code(self) -> str:
        """Returns the new file contend or None if there are no replacepents to
        apply."""
        replacements = list(self.replacements)
        replacements.sort()

        self._check()

        code = self.filename.read_text("utf-8")

        format_whole_file = enforce_formatting() or code == format_code(
            code, self.filename
        )

        if not format_whole_file:
            logging.info(f"file is not formatted with black: {self.filename}")
            import black

            logging.info(f"black version: {black.__version__}")

        line_numbers = LineNumbers(code)

        new_code = asttokens.util.replace(
            code,
            [
                (
                    r.range.start.offset(line_numbers),
                    r.range.end.offset(line_numbers),
                    r.text,
                )
                for r in replacements
            ],
        )

        if format_whole_file:
            new_code = format_code(new_code, self.filename)

        return new_code

    def diff(self):
        return "\n".join(
            islice(
                unified_diff(self.source.splitlines(), self.new_code().splitlines()),
                2,
                None,
            )
        ).strip()


class ChangeRecorder:

    def __init__(self):
        self._source_files = defaultdict(SourceFile)
        self._changes = []

    def get_source(self, filename) -> SourceFile:
        filename = pathlib.Path(filename)
        if filename not in self._source_files:
            self._source_files[filename] = SourceFile(filename)

        return self._source_files[filename]

    def files(self) -> Iterable[SourceFile]:
        return self._source_files.values()

    def new_change(self):
        return Change(self)

    def num_fixes(self):
        changes = set()
        for file in self._source_files.values():
            changes.update(change.change_id for change in file.replacements)
        return len(changes)

    def fix_all(self):
        for file in self._source_files.values():
            file.rewrite()

    def virtual_write(self):
        for file in self._source_files.values():
            file.virtual_write()

    def dump(self):  # pragma: no cover
        for file in self._source_files.values():
            print("file:", file.filename)
            for change in file.replacements:
                print("  change:", change)



from __future__ import annotations

import ast
import warnings
from abc import ABC
from collections import defaultdict
from dataclasses import MISSING
from dataclasses import fields
from dataclasses import is_dataclass
from typing import Any

from .._change import CallArg
from .._change import Delete
from ..syntax_warnings import InlineSnapshotSyntaxWarning
from .adapter import Adapter
from .adapter import Item
from .adapter import adapter_map


def get_adapter_for_type(value_type):
    subclasses = GenericCallAdapter.__subclasses__()
    options = [cls for cls in subclasses if cls.check_type(value_type)]

    if not options:
        return

    assert len(options) == 1
    return options[0]


class Argument:
    value: Any
    is_default: bool = False

    def __init__(self, value, is_default=False):
        self.value = value
        self.is_default = is_default


class GenericCallAdapter(Adapter):

    @classmethod
    def check_type(cls, value_type) -> bool:
        raise NotImplementedError(cls)

    @classmethod
    def arguments(cls, value) -> tuple[list[Argument], dict[str, Argument]]:
        raise NotImplementedError(cls)

    @classmethod
    def argument(cls, value, pos_or_name) -> Any:
        raise NotImplementedError(cls)

    @classmethod
    def repr(cls, value):

        args, kwargs = cls.arguments(value)

        arguments = [repr(value.value) for value in args] + [
            f"{key}={repr(value.value)}"
            for key, value in kwargs.items()
            if not value.is_default
        ]

        return f"{repr(type(value))}({', '.join(arguments)})"

    @classmethod
    def map(cls, value, map_function):
        new_args, new_kwargs = cls.arguments(value)
        return type(value)(
            *[adapter_map(arg.value, map_function) for arg in new_args],
            **{
                k: adapter_map(kwarg.value, map_function)
                for k, kwarg in new_kwargs.items()
            },
        )

    @classmethod
    def items(cls, value, node):
        new_args, new_kwargs = cls.arguments(value)

        if node is not None:
            assert isinstance(node, ast.Call)
            assert all(kw.arg for kw in node.keywords)
            kw_arg_node = {kw.arg: kw.value for kw in node.keywords if kw.arg}.get

            def pos_arg_node(pos):
                return node.args[pos]

        else:

            def kw_arg_node(_):
                return None

            def pos_arg_node(_):
                return None

        return [
            Item(value=arg.value, node=pos_arg_node(i))
            for i, arg in enumerate(new_args)
        ] + [
            Item(value=kw.value, node=kw_arg_node(name))
            for name, kw in new_kwargs.items()
        ]

    def assign(self, old_value, old_node, new_value):
        if old_node is None or not isinstance(old_node, ast.Call):
            result = yield from self.value_assign(old_value, old_node, new_value)
            return result

        call_type = self.context.eval(old_node.func)

        if not (isinstance(call_type, type) and self.check_type(call_type)):
            result = yield from self.value_assign(old_value, old_node, new_value)
            return result

        # positional arguments
        for pos_arg in old_node.args:
            if isinstance(pos_arg, ast.Starred):
                warnings.warn_explicit(
                    "star-expressions are not supported inside snapshots",
                    filename=self.context.file._source.filename,
                    lineno=pos_arg.lineno,
                    category=InlineSnapshotSyntaxWarning,
                )
                return old_value

        # keyword arguments
        for kw in old_node.keywords:
            if kw.arg is None:
                warnings.warn_explicit(
                    "star-expressions are not supported inside snapshots",
                    filename=self.context.file._source.filename,
                    lineno=kw.value.lineno,
                    category=InlineSnapshotSyntaxWarning,
                )
                return old_value

        new_args, new_kwargs = self.arguments(new_value)

        # positional arguments

        result_args = []

        for i, (new_value_element, node) in enumerate(zip(new_args, old_node.args)):
            old_value_element = self.argument(old_value, i)
            result = yield from self.get_adapter(
                old_value_element, new_value_element.value
            ).assign(old_value_element, node, new_value_element.value)
            result_args.append(result)

        if len(old_node.args) > len(new_args):
            for arg_pos, node in list(enumerate(old_node.args))[len(new_args) :]:
                yield Delete(
                    "fix",
                    self.context.file._source,
                    node,
                    self.argument(old_value, arg_pos),
                )

        if len(old_node.args) < len(new_args):
            for insert_pos, value in list(enumerate(new_args))[len(old_node.args) :]:
                yield CallArg(
                    flag="fix",
                    file=self.context.file._source,
                    node=old_node,
                    arg_pos=insert_pos,
                    arg_name=None,
                    new_code=self.context.file._value_to_code(value.value),
                    new_value=value.value,
                )

        # keyword arguments
        result_kwargs = {}
        for kw in old_node.keywords:
            if kw.arg not in new_kwargs or new_kwargs[kw.arg].is_default:
                # delete entries
                yield Delete(
                    (
                        "update"
                        if self.argument(old_value, kw.arg)
                        == self.argument(new_value, kw.arg)
                        else "fix"
                    ),
                    self.context.file._source,
                    kw.value,
                    self.argument(old_value, kw.arg),
                )

        old_node_kwargs = {kw.arg: kw.value for kw in old_node.keywords}

        to_insert = []
        insert_pos = 0
        for key, new_value_element in new_kwargs.items():
            if new_value_element.is_default:
                continue
            if key not in old_node_kwargs:
                # add new values
                to_insert.append((key, new_value_element.value))
                result_kwargs[key] = new_value_element.value
            else:
                node = old_node_kwargs[key]

                # check values with same keys
                old_value_element = self.argument(old_value, key)
                result_kwargs[key] = yield from self.get_adapter(
                    old_value_element, new_value_element.value
                ).assign(old_value_element, node, new_value_element.value)

                if to_insert:
                    for key, value in to_insert:

                        yield CallArg(
                            flag="fix",
                            file=self.context.file._source,
                            node=old_node,
                            arg_pos=insert_pos,
                            arg_name=key,
                            new_code=self.context.file._value_to_code(value),
                            new_value=value,
                        )
                    to_insert = []

                insert_pos += 1

        if to_insert:

            for key, value in to_insert:

                yield CallArg(
                    flag="fix",
                    file=self.context.file._source,
                    node=old_node,
                    arg_pos=insert_pos,
                    arg_name=key,
                    new_code=self.context.file._value_to_code(value),
                    new_value=value,
                )

        return type(old_value)(*result_args, **result_kwargs)


class DataclassAdapter(GenericCallAdapter):

    @classmethod
    def check_type(cls, value):
        return is_dataclass(value)

    @classmethod
    def arguments(cls, value):

        kwargs = {}

        for field in fields(value):  # type: ignore
            if field.repr:
                field_value = getattr(value, field.name)
                is_default = False

                if field.default != MISSING and field.default == field_value:
                    is_default = True

                if (
                    field.default_factory != MISSING
                    and field.default_factory() == field_value
                ):
                    is_default = True

                kwargs[field.name] = Argument(value=field_value, is_default=is_default)

        return ([], kwargs)

    def argument(self, value, pos_or_name):
        if isinstance(pos_or_name, str):
            return getattr(value, pos_or_name)
        else:
            args = [field for field in fields(value) if field.init]
            return args[pos_or_name]


try:
    import attrs
except ImportError:  # pragma: no cover
    pass
else:

    class AttrAdapter(GenericCallAdapter):

        @classmethod
        def check_type(cls, value):
            return attrs.has(value)

        @classmethod
        def arguments(cls, value):

            kwargs = {}

            for field in attrs.fields(type(value)):
                if field.repr:
                    field_value = getattr(value, field.name)
                    is_default = False

                    if field.default is not attrs.NOTHING:

                        default_value = (
                            field.default
                            if not isinstance(field.default, attrs.Factory)
                            else (
                                field.default.factory()
                                if not field.default.takes_self
                                else field.default.factory(value)
                            )
                        )

                        if default_value == field_value:

                            is_default = True

                    kwargs[field.name] = Argument(
                        value=field_value, is_default=is_default
                    )

            return ([], kwargs)

        def argument(self, value, pos_or_name):
            assert isinstance(pos_or_name, str)
            return getattr(value, pos_or_name)


try:
    import pydantic
except ImportError:  # pragma: no cover
    pass
else:
    # import pydantic
    if pydantic.version.VERSION.startswith("1."):
        # pydantic v1
        from pydantic.fields import Undefined as PydanticUndefined  # type: ignore[attr-defined,no-redef]

        def get_fields(value):
            return value.__fields__

    else:
        # pydantic v2
        from pydantic_core import PydanticUndefined

        def get_fields(value):
            return type(value).model_fields

    from pydantic import BaseModel

    class PydanticContainer(GenericCallAdapter):

        @classmethod
        def check_type(cls, value):
            return issubclass(value, BaseModel)

        @classmethod
        def arguments(cls, value):

            kwargs = {}

            for name, field in get_fields(value).items():  # type: ignore
                if getattr(field, "repr", True):
                    field_value = getattr(value, name)
                    is_default = False

                    if (
                        field.default is not PydanticUndefined
                        and field.default == field_value
                    ):
                        is_default = True

                    if (
                        field.default_factory is not None
                        and field.default_factory() == field_value
                    ):
                        is_default = True

                    kwargs[name] = Argument(value=field_value, is_default=is_default)

            return ([], kwargs)

        @classmethod
        def argument(cls, value, pos_or_name):
            assert isinstance(pos_or_name, str)
            return getattr(value, pos_or_name)


class IsNamedTuple(ABC):
    _inline_snapshot_name = "namedtuple"

    _fields: tuple
    _field_defaults: dict

    @classmethod
    def __subclasshook__(cls, t):
        b = t.__bases__
        if len(b) != 1 or b[0] != tuple:
            return False
        f = getattr(t, "_fields", None)
        if not isinstance(f, tuple):
            return False
        return all(type(n) == str for n in f)


class NamedTupleAdapter(GenericCallAdapter):

    @classmethod
    def check_type(cls, value):
        return issubclass(value, IsNamedTuple)

    @classmethod
    def arguments(cls, value: IsNamedTuple):

        return (
            [],
            {
                field: Argument(value=getattr(value, field))
                for field in value._fields
                if field not in value._field_defaults
                or getattr(value, field) != value._field_defaults[field]
            },
        )

    def argument(self, value, pos_or_name):
        assert isinstance(pos_or_name, str)
        return getattr(value, pos_or_name)


class DefaultDictAdapter(GenericCallAdapter):
    @classmethod
    def check_type(cls, value):
        return issubclass(value, defaultdict)

    @classmethod
    def arguments(cls, value: defaultdict):

        return (
            [Argument(value=value.default_factory), Argument(value=dict(value))],
            {},
        )

    def argument(self, value, pos_or_name):
        assert isinstance(pos_or_name, int)
        if pos_or_name == 0:
            return value.default_factory
        elif pos_or_name == 1:
            return dict(value)
        assert False



"""The following functions are build on top of inline-snapshot and could also
be implemented in an extra library.

They are part of inline-snapshot because they are general useful and do
not depend on other libraries.
"""

import contextlib
import io
import warnings
from contextlib import redirect_stderr
from contextlib import redirect_stdout
from typing import List
from typing import Tuple
from typing import Union

from inline_snapshot._types import Snapshot


@contextlib.contextmanager
def raises(exception: Snapshot[str]):
    """Check that an exception is raised.

    Parameters:
        exception: snapshot which is compared with `#!python f"{type}: {message}"` if an exception occurred or `#!python "<no exception>"` if no exception was raised.

    === "original"

        <!-- inline-snapshot: first_block outcome-passed=1 outcome-errors=1 -->
        ``` python
        from inline_snapshot import snapshot
        from inline_snapshot.extra import raises


        def test_raises():
            with raises(snapshot()):
                1 / 0
        ```

    === "--inline-snapshot=create"

        <!-- inline-snapshot: create outcome-passed=1 outcome-errors=1 -->
        ``` python hl_lines="6"
        from inline_snapshot import snapshot
        from inline_snapshot.extra import raises


        def test_raises():
            with raises(snapshot("ZeroDivisionError: division by zero")):
                1 / 0
        ```
    """

    try:
        yield
    except Exception as ex:
        msg = str(ex)
        if "\n" in msg:
            assert f"{type(ex).__name__}:\n{ex}" == exception
        else:
            assert f"{type(ex).__name__}: {ex}" == exception
    else:
        assert "<no exception>" == exception


@contextlib.contextmanager
def prints(*, stdout: Snapshot[str] = "", stderr: Snapshot[str] = ""):
    """Uses `contextlib.redirect_stderr/stdout` to capture the output and
    compare it with the snapshots. `dirty_equals.IsStr` can be used to ignore
    the output if needed.

    Parameters:
        stdout: snapshot which is compared to the recorded output
        stderr: snapshot which is compared to the recorded error output

    === "original"

        <!-- inline-snapshot: first_block outcome-passed=1 outcome-errors=1 -->
        ``` python
        from inline_snapshot import snapshot
        from inline_snapshot.extra import prints
        import sys


        def test_prints():
            with prints(stdout=snapshot(), stderr=snapshot()):
                print("hello world")
                print("some error", file=sys.stderr)
        ```

    === "--inline-snapshot=create"

        <!-- inline-snapshot: create outcome-passed=1 outcome-errors=1 -->
        ``` python hl_lines="7 8 9"
        from inline_snapshot import snapshot
        from inline_snapshot.extra import prints
        import sys


        def test_prints():
            with prints(
                stdout=snapshot("hello world\\n"), stderr=snapshot("some error\\n")
            ):
                print("hello world")
                print("some error", file=sys.stderr)
        ```

    === "ignore stdout"

        <!-- inline-snapshot: outcome-passed=1 -->
        ``` python hl_lines="3 9 10"
        from inline_snapshot import snapshot
        from inline_snapshot.extra import prints
        from dirty_equals import IsStr
        import sys


        def test_prints():
            with prints(
                stdout=IsStr(),
                stderr=snapshot("some error\\n"),
            ):
                print("hello world")
                print("some error", file=sys.stderr)
        ```
    """

    with redirect_stdout(io.StringIO()) as stdout_io:
        with redirect_stderr(io.StringIO()) as stderr_io:
            yield

    assert stderr_io.getvalue() == stderr
    assert stdout_io.getvalue() == stdout


Warning = Union[str, Tuple[int, str], Tuple[str, str], Tuple[str, int, str]]


@contextlib.contextmanager
def warns(
    expected_warnings: Snapshot[List[Warning]],
    /,
    include_line: bool = False,
    include_file: bool = False,
):
    """
    Captures warnings with `warnings.catch_warnings` and compares them against expected warnings.

    Parameters:
        expected_warnings: Snapshot containing a list of expected warnings.
        include_line: If `True`, each expected warning is a tuple `(linenumber, message)`.
        include_file: If `True`, each expected warning is a tuple `(filename, message)`.

    The format of the expected warning:

    - `(filename, linenumber, message)` if both `include_line` and `include_file` are `True`.
    - `(linenumber, message)` if only `include_line` is `True`.
    - `(filename, message)` if only `include_file` is `True`.
    - A string `message` if both are `False`.

    === "original"

        <!-- inline-snapshot: first_block outcome-passed=1 outcome-errors=1 -->
        ``` python
        from inline_snapshot import snapshot
        from inline_snapshot.extra import warns
        from warnings import warn


        def test_warns():
            with warns(snapshot(), include_line=True):
                warn("some problem")
        ```

    === "--inline-snapshot=create"

        <!-- inline-snapshot: create fix outcome-passed=1 outcome-errors=1 -->
        ``` python hl_lines="7"
        from inline_snapshot import snapshot
        from inline_snapshot.extra import warns
        from warnings import warn


        def test_warns():
            with warns(snapshot([(8, "UserWarning: some problem")]), include_line=True):
                warn("some problem")
        ```
    """
    with warnings.catch_warnings(record=True) as result:
        warnings.simplefilter("always")
        yield

    def make_warning(w):
        message = f"{w.category.__name__}: {w.message}"
        if not include_line and not include_file:
            return message
        message = (message,)

        if include_line:
            message = (w.lineno, *message)
        if include_file:
            message = (w.filename, *message)

        return message

    assert [make_warning(w) for w in result] == expected_warnings



from __future__ import annotations

import ast
import warnings

from .._change import Replace
from .._code_repr import value_code_repr
from .._sentinels import undefined
from .._unmanaged import Unmanaged
from .._unmanaged import update_allowed
from .._utils import value_to_token
from ..syntax_warnings import InlineSnapshotInfo
from .adapter import Adapter


class ValueAdapter(Adapter):

    @classmethod
    def repr(cls, value):
        return value_code_repr(value)

    @classmethod
    def map(cls, value, map_function):
        return map_function(value)

    def assign(self, old_value, old_node, new_value):
        # generic fallback

        # because IsStr() != IsStr()
        if isinstance(old_value, Unmanaged):
            return old_value

        if old_node is None:
            new_token = []
        else:
            new_token = value_to_token(new_value)

        if isinstance(old_node, ast.JoinedStr) and isinstance(new_value, str):
            if not old_value == new_value:
                warnings.warn_explicit(
                    f"inline-snapshot will be able to fix f-strings in the future.\nThe current string value is:\n   {new_value!r}",
                    filename=self.context.file._source.filename,
                    lineno=old_node.lineno,
                    category=InlineSnapshotInfo,
                )
            return old_value

        if not old_value == new_value:
            if old_value is undefined:
                flag = "create"
            else:
                flag = "fix"
        elif (
            old_node is not None
            and update_allowed(old_value)
            and self.context.file._token_of_node(old_node) != new_token
        ):
            flag = "update"
        else:
            # equal and equal repr
            return old_value

        new_code = self.context.file._token_to_code(new_token)

        yield Replace(
            node=old_node,
            file=self.context.file._source,
            new_code=new_code,
            flag=flag,
            old_value=old_value,
            new_value=new_value,
        )

        return new_value



from contextlib import contextmanager


def compare_only():
    return _eq_check_only


_eq_check_only = False


@contextmanager
def compare_context():
    global _eq_check_only
    old_eq_only = _eq_check_only
    _eq_check_only = True
    yield
    _eq_check_only = old_eq_only



from __future__ import annotations

import contextlib
from dataclasses import dataclass
from dataclasses import field
from typing import TYPE_CHECKING
from typing import Generator

from ._flags import Flags

if TYPE_CHECKING:
    from ._external import DiscStorage


@dataclass
class State:
    # snapshot
    missing_values: int = 0
    incorrect_values: int = 0

    snapshots: dict = field(default_factory=dict)
    update_flags: Flags = field(default_factory=Flags)
    active: bool = True
    files_with_snapshots: set[str] = field(default_factory=set)

    # external
    storage: DiscStorage | None = None

    flags: set[str] = field(default_factory=set)


_latest_global_states: list[State] = []

_current: State = State()
_current.active = False


def state() -> State:
    global _current
    return _current


def enter_snapshot_context():
    global _current
    _latest_global_states.append(_current)
    _current = State()


def leave_snapshot_context():
    global _current
    _current = _latest_global_states.pop()


@contextlib.contextmanager
def snapshot_env() -> Generator[State]:

    enter_snapshot_context()

    try:
        yield _current
    finally:
        leave_snapshot_context()



from typing import IO
from typing import Any
from typing import Set

from inline_snapshot._is import Is
from inline_snapshot._snapshot.generic_value import GenericValue
from inline_snapshot._unmanaged import Unmanaged


def fix_pytest_diff():
    from _pytest._io.pprint import PrettyPrinter

    def _pprint_snapshot(
        self,
        object: Any,
        stream: IO[str],
        indent: int,
        allowance: int,
        context: Set[int],
        level: int,
    ) -> None:
        self._format(object._old_value, stream, indent, allowance, context, level)

    PrettyPrinter._dispatch[GenericValue.__repr__] = _pprint_snapshot

    def _pprint_unmanaged(
        self,
        object: Any,
        stream: IO[str],
        indent: int,
        allowance: int,
        context: Set[int],
        level: int,
    ) -> None:
        self._format(object.value, stream, indent, allowance, context, level)

    PrettyPrinter._dispatch[Unmanaged.__repr__] = _pprint_unmanaged

    def _pprint_is(
        self,
        object: Any,
        stream: IO[str],
        indent: int,
        allowance: int,
        context: Set[int],
        level: int,
    ) -> None:
        self._format(object.value, stream, indent, allowance, context, level)

    PrettyPrinter._dispatch[Is.__repr__] = _pprint_is



from __future__ import annotations

import os
import platform
import re
import subprocess as sp
import sys
import traceback
from argparse import ArgumentParser
from io import StringIO
from pathlib import Path
from pathlib import PurePosixPath
from tempfile import TemporaryDirectory
from typing import Any

from rich.console import Console

from inline_snapshot._exceptions import UsageError
from inline_snapshot._external import DiscStorage
from inline_snapshot._problems import report_problems

from .._change import apply_all
from .._flags import Flags
from .._global_state import snapshot_env
from .._rewrite_code import ChangeRecorder
from .._types import Category
from .._types import Snapshot

ansi_escape = re.compile(
    r"""
    \x1B  # ESC
    (?:   # 7-bit C1 Fe (except CSI)
        [@-Z\\-_]
    |     # or [ for CSI, followed by a control sequence
        \[
        [0-?]*  # Parameter bytes
        [ -/]*  # Intermediate bytes
        [@-~]   # Final byte
    )
""",
    re.VERBOSE,
)


def normalize(text):
    text = ansi_escape.sub("", text)

    # fix windows problems
    text = text.replace("\u2500", "-")
    text = text.replace("\r", "")
    return text


# this code is copied from pytest

# Regex to match the session duration string in the summary: "74.34s".
rex_session_duration = re.compile(r"\d+\.\d\ds")
# Regex to match all the counts and phrases in the summary line: "34 passed, 111 skipped".
rex_outcome = re.compile(r"(\d+) (\w+)")


def parse_outcomes(lines):
    for line in reversed(lines):
        if rex_session_duration.search(line):
            outcomes = rex_outcome.findall(line)
            ret = {noun: int(count) for (count, noun) in outcomes}
            break
        else:
            pass  # pragma: no cover
    else:
        raise ValueError("Pytest terminal summary report not found")  # pragma: no cover

    to_plural = {
        "warning": "warnings",
        "error": "errors",
    }
    return {to_plural.get(k, k): v for k, v in ret.items()}


class Example:
    def __init__(self, files: str | dict[str, str]):
        """
        Parameters:
            files: a collection of files where inline-snapshot operates on,
                   or just a string which will be saved as *test_something.py*.
        """
        if isinstance(files, str):
            files = {"test_something.py": files}

        self.files = files

        self.dump_files()

    def dump_files(self):
        for name, content in self.files.items():
            print(f"file: {name}")
            print(content)
            print()

    def _write_files(self, dir: Path):
        for name, content in self.files.items():
            filename = dir / name
            filename.parent.mkdir(exist_ok=True, parents=True)
            filename.write_text(content)

    def _read_files(self, dir: Path):
        storage_dir = dir / ".inline-snapshot"

        return {
            str(p.relative_to(dir)): p.read_text()
            for p in [*dir.iterdir(), *dir.rglob("*.py"), *storage_dir.rglob("*")]
            if p.is_file()
        }

    def with_files(self, extra_files):
        return Example(self.files | extra_files)

    def code_change(self, src, dest):
        return Example(
            {name: file.replace(src, dest) for name, file in self.files.items()}
        )

    def run_inline(
        self,
        args: list[str] = [],
        *,
        reported_categories: Snapshot[list[Category]] | None = None,
        changed_files: Snapshot[dict[str, str]] | None = None,
        report: Snapshot[str] | None = None,
        raises: Snapshot[str] | None = None,
    ) -> Example:
        """Execute the example files in process and run every `test_*`
        function.

        This is useful for fast test execution.

        Parameters:
            args: inline-snapshot arguments (supports only "--inline-snapshot=fix|create|..." ).
            reported_categories: snapshot of categories which inline-snapshot thinks could be applied.
            changed_files: snapshot of files which are changed by this run.
            raises: snapshot of the exception which is raised during the test execution.
                    It is required if your code raises an exception.

        Returns:
            A new Example instance which contains the changed files.
        """

        parser = ArgumentParser()

        parser.add_argument(
            "--inline-snapshot",
            metavar="(disable,short-report,report,review,create,update,trim,fix)*",
            dest="inline_snapshot",
            help="update specific snapshot values:\n"
            "disable: disable the snapshot logic\n"
            "short-report: show a short report\n"
            "report: show a longer report with a diff report\n"
            "review: allow to approve the changes interactive\n"
            "create: creates snapshots which are currently not defined\n"
            "update: update snapshots even if they are defined\n"
            "trim: changes the snapshot in a way which will make the snapshot more precise.\n"
            "fix: change snapshots which currently break your tests\n",
        )
        parsed_args = parser.parse_args(args)
        flags = (parsed_args.inline_snapshot or "").split(",")

        with TemporaryDirectory() as dir:
            tmp_path = Path(dir)

            self._write_files(tmp_path)

            raised_exception = []
            with snapshot_env() as state:
                recorder = ChangeRecorder()
                state.update_flags = Flags({*flags})
                state.storage = DiscStorage(tmp_path / ".storage")
                try:
                    tests_found = False
                    for filename in tmp_path.glob("*.py"):
                        globals: dict[str, Any] = {}
                        print("run> pytest", filename)
                        exec(
                            compile(filename.read_text("utf-8"), filename, "exec"),
                            globals,
                        )

                        # run all test_* functions
                        tests = [
                            v
                            for k, v in globals.items()
                            if (k.startswith("test_") or k == "test") and callable(v)
                        ]
                        tests_found |= len(tests) != 0

                        for v in tests:
                            try:
                                v()
                            except Exception as e:
                                traceback.print_exc()
                                raised_exception.append(e)

                    if not tests_found:
                        raise UsageError("no test_*() functions in the example")
                finally:
                    state.active = False

                changes = []
                for snapshot in state.snapshots.values():
                    changes += snapshot._changes()

                snapshot_flags = {change.flag for change in changes}

                apply_all(
                    [
                        change
                        for change in changes
                        if change.flag in state.update_flags.to_set()
                    ],
                    recorder,
                )
                recorder.fix_all()

                report_output = StringIO()
                console = Console(file=report_output, width=80)

                # TODO: add all the report output here
                report_problems(lambda: console)

            if reported_categories is not None:
                assert sorted(snapshot_flags) == reported_categories

            if raised_exception:
                if raises is None:
                    raise raised_exception[0]

                assert raises == "\n".join(
                    f"{type(e).__name__}:\n" + str(e) for e in raised_exception
                )
            else:
                assert raises == None

            if changed_files is not None:
                current_files = {}

                for name, content in sorted(self._read_files(tmp_path).items()):
                    if name not in self.files or self.files[name] != content:
                        current_files[name] = content
                assert changed_files == current_files

            if report is not None:
                assert report == normalize(report_output.getvalue())

            return Example(self._read_files(tmp_path))

    def run_pytest(
        self,
        args: list[str] = [],
        *,
        term_columns=80,
        env: dict[str, str] = {},
        changed_files: Snapshot[dict[str, str]] | None = None,
        report: Snapshot[str] | None = None,
        error: Snapshot[str] | None = None,
        stderr: Snapshot[str] | None = None,
        returncode: Snapshot[int] = 0,
        stdin: bytes = b"",
        outcomes: Snapshot[dict[str, int]] | None = None,
    ) -> Example:
        """Run pytest with the given args and env variables in an separate
        process.

        It can be used to test the interaction between your code and pytest, but it is a bit slower than `run_inline`

        Parameters:
            args: pytest arguments like "--inline-snapshot=fix"
            env: dict of environment variables
            changed_files: snapshot of files which are changed by this run.
            report: snapshot of the report at the end of the pytest run.
            stderr: pytest stderr output
            returncode: snapshot of the pytest returncode.

        Returns:
            A new Example instance which contains the changed files.
        """

        with TemporaryDirectory() as dir:
            tmp_path = Path(dir)
            self._write_files(tmp_path)

            cmd = [sys.executable, "-m", "pytest", *args]

            command_env = dict(os.environ)
            command_env["TERM"] = "unknown"
            command_env["COLUMNS"] = str(
                term_columns + 1 if platform.system() == "Windows" else term_columns
            )
            command_env.pop("CI", None)
            command_env.pop("GITHUB_ACTIONS", None)

            if stdin:
                # makes Console.is_terminal == True
                command_env["FORCE_COLOR"] = "true"

            command_env.update(env)

            result = sp.run(
                cmd, cwd=tmp_path, capture_output=True, env=command_env, input=stdin
            )

            result_stdout = result.stdout.decode()
            result_stderr = result.stderr.decode()
            result_returncode = result.returncode

            print("run>", *cmd)
            print("stdout:")
            print(result_stdout)
            print("stderr:")
            print(result_stderr)

            assert result.returncode == returncode

            if stderr is not None:

                original = result_stderr.splitlines()
                lines = [
                    line
                    for line in original
                    if not any(
                        s in line
                        for s in [
                            'No entry for terminal type "unknown"',
                            "using dumb terminal settings.",
                        ]
                    )
                ]

                assert "\n".join(lines) == stderr

            if report is not None:

                report_list = []
                record = False
                for line in result_stdout.splitlines():
                    line = normalize(line.strip())
                    if line.startswith("===="):
                        record = False

                    if record and line:
                        report_list.append(line)

                    if (
                        line.startswith(("-----", "═════"))
                        and "inline-snapshot" in line
                    ):
                        record = True

                report_str = "\n".join(report_list)

                assert report_str == report, repr(report_str)

            if error is not None:
                assert (
                    error
                    == "\n".join(
                        [
                            line
                            for line in result_stdout.splitlines()
                            if line and line[:2] in ("> ", "E ")
                        ]
                    )
                    + "\n"
                )

            if changed_files is not None:
                current_files = {}

                for name, content in sorted(self._read_files(tmp_path).items()):
                    if name not in self.files or self.files[name] != content:
                        current_files[str(PurePosixPath(*Path(name).parts))] = content
                assert changed_files == current_files

            if outcomes is not None:
                assert outcomes == parse_outcomes(result_stdout.splitlines())

            return Example(self._read_files(tmp_path))



import sys
from dataclasses import dataclass
from dataclasses import field
from pathlib import Path
from typing import Dict
from typing import List
from typing import Optional

if sys.version_info >= (3, 11):
    from tomllib import loads
else:
    from tomli import loads


@dataclass
class Config:
    hash_length: int = 12
    default_flags: List[str] = field(default_factory=lambda: ["short-report"])
    default_flags_tui: List[str] = field(default_factory=lambda: ["short-report"])
    shortcuts: Dict[str, List[str]] = field(default_factory=dict)
    format_command: str = ""
    storage_dir: Optional[Path] = None
    show_updates: bool = False


config = Config()


def read_config(path: Path, config=Config()) -> Config:
    tool_config = {}

    if path.exists():
        data = loads(path.read_text("utf-8"))

        try:
            tool_config = data["tool"]["inline-snapshot"]
        except KeyError:
            pass

    try:
        config.hash_length = tool_config["hash-length"]
    except KeyError:
        pass

    try:
        config.default_flags = tool_config["default-flags"]
    except KeyError:
        pass

    try:
        config.default_flags_tui = tool_config["default-flags-tui"]
    except KeyError:
        pass

    try:
        config.show_updates = tool_config["show-updates"]
    except KeyError:
        pass

    config.shortcuts = tool_config.get(
        "shortcuts", {"fix": ["create", "fix"], "review": ["review"]}
    )

    if storage_dir := tool_config.get("storage-dir"):
        storage_dir = Path(storage_dir)
        if not storage_dir.is_absolute():
            # Make it relative to pyproject.toml, and absolute.
            storage_dir = path.parent.joinpath(storage_dir).absolute()
        config.storage_dir = storage_dir

    config.format_command = tool_config.get("format-command", "")

    return config



from typing import Iterator

from inline_snapshot._adapter.adapter import adapter_map

from .._adapter.adapter import AdapterContext
from .._adapter.adapter import get_adapter_type
from .._change import Change
from .._change import Replace
from .._sentinels import undefined
from .._unmanaged import Unmanaged
from .._unmanaged import map_unmanaged
from .._utils import value_to_token
from .generic_value import GenericValue


class UndecidedValue(GenericValue):
    def __init__(self, old_value, ast_node, context: AdapterContext):

        old_value = adapter_map(old_value, map_unmanaged)
        self._old_value = old_value
        self._new_value = undefined
        self._ast_node = ast_node
        self._context = context

    def _change(self, cls):
        self.__class__ = cls

    def _new_code(self):
        assert False

    def _get_changes(self) -> Iterator[Change]:

        def handle(node, obj):

            adapter = get_adapter_type(obj)
            if adapter is not None and hasattr(adapter, "items"):
                for item in adapter.items(obj, node):
                    yield from handle(item.node, item.value)
                return

            if not isinstance(obj, Unmanaged) and node is not None:
                new_token = value_to_token(obj)
                if self._file._token_of_node(node) != new_token:
                    new_code = self._file._token_to_code(new_token)

                    yield Replace(
                        node=self._ast_node,
                        file=self._file,
                        new_code=new_code,
                        flag="update",
                        old_value=self._old_value,
                        new_value=self._old_value,
                    )

        if self._file._source is not None:
            yield from handle(self._ast_node, self._old_value)

    # functions which determine the type

    def __eq__(self, other):
        from .._snapshot.eq_value import EqValue

        self._change(EqValue)
        return self == other

    def __le__(self, other):
        from .._snapshot.min_max_value import MinValue

        self._change(MinValue)
        return self <= other

    def __ge__(self, other):
        from .._snapshot.min_max_value import MaxValue

        self._change(MaxValue)
        return self >= other

    def __contains__(self, other):
        from .._snapshot.collection_value import CollectionValue

        self._change(CollectionValue)
        return other in self

    def __getitem__(self, item):
        from .._snapshot.dict_value import DictValue

        self._change(DictValue)
        return self[item]



import ast
from collections import defaultdict
from dataclasses import dataclass
from typing import Any
from typing import DefaultDict
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union
from typing import cast

from asttokens.util import Token
from executing.executing import EnhancedAST

from inline_snapshot._source_file import SourceFile

from ._rewrite_code import ChangeRecorder
from ._rewrite_code import end_of
from ._rewrite_code import start_of


@dataclass()
class Change:
    flag: str
    file: SourceFile

    @property
    def filename(self):
        return self.file.filename

    def apply(self, recorder: ChangeRecorder):
        raise NotImplementedError()


@dataclass()
class Delete(Change):
    node: ast.AST
    old_value: Any


@dataclass()
class AddArgument(Change):
    node: ast.Call

    position: Optional[int]
    name: Optional[str]

    new_code: str
    new_value: Any


@dataclass()
class ListInsert(Change):
    node: ast.List
    position: int

    new_code: List[str]
    new_values: List[Any]


@dataclass()
class DictInsert(Change):
    node: ast.Dict
    position: int

    new_code: List[Tuple[str, str]]
    new_values: List[Tuple[Any, Any]]


@dataclass()
class Replace(Change):
    node: ast.AST

    new_code: str
    old_value: Any
    new_value: Any

    def apply(self, recorder: ChangeRecorder):
        change = recorder.new_change()
        range = self.file.asttokens().get_text_positions(self.node, False)
        change.replace(range, self.new_code, filename=self.filename)


@dataclass()
class CallArg(Change):
    node: Optional[ast.Call]
    arg_pos: Optional[int]
    arg_name: Optional[str]

    new_code: str
    new_value: Any


TokenRange = Tuple[Token, Token]


def brace_tokens(source, node) -> TokenRange:
    first_token, *_, end_token = source.asttokens().get_tokens(node)
    return first_token, end_token


def generic_sequence_update(
    source: SourceFile,
    parent: Union[ast.List, ast.Tuple, ast.Dict, ast.Call],
    brace_tokens: TokenRange,
    parent_elements: List[Union[TokenRange, None]],
    to_insert: Dict[int, List[str]],
    recorder: ChangeRecorder,
):
    rec = recorder.new_change()

    new_code = []
    deleted = False
    last_token, end_token = brace_tokens
    is_start = True
    elements = 0

    for index, entry in enumerate(parent_elements):
        if index in to_insert:
            new_code += to_insert[index]
        if entry is None:
            deleted = True
        else:
            first_token, new_last_token = entry
            elements += len(new_code) + 1

            if deleted or new_code:

                code = ""
                if new_code:
                    code = ", ".join(new_code) + ", "
                if not is_start:
                    code = ", " + code

                rec.replace(
                    (end_of(last_token), start_of(first_token)),
                    code,
                    filename=source.filename,
                )
            new_code = []
            deleted = False
            last_token = new_last_token
            is_start = False

    if len(parent_elements) in to_insert:
        new_code += to_insert[len(parent_elements)]
        elements += len(new_code)

    if new_code or deleted or elements == 1 or len(parent_elements) <= 1:
        code = ", ".join(new_code)
        if not is_start and code:
            code = ", " + code

        if elements == 1 and isinstance(parent, ast.Tuple):
            # trailing comma for tuples (1,)
            code += ","

        rec.replace(
            (end_of(last_token), start_of(end_token)),
            code,
            filename=source.filename,
        )


def apply_all(all_changes: List[Change], recorder: ChangeRecorder):
    by_parent: Dict[
        EnhancedAST, List[Union[Delete, DictInsert, ListInsert, CallArg]]
    ] = defaultdict(list)
    sources: Dict[EnhancedAST, SourceFile] = {}

    for change in all_changes:
        if isinstance(change, Delete):
            node = cast(EnhancedAST, change.node).parent
            if isinstance(node, ast.keyword):
                node = node.parent
            by_parent[node].append(change)
            sources[node] = change.file

        elif isinstance(change, (DictInsert, ListInsert, CallArg)):
            node = cast(EnhancedAST, change.node)
            by_parent[node].append(change)
            sources[node] = change.file
        else:
            change.apply(recorder)

    for parent, changes in by_parent.items():
        source = sources[parent]

        if isinstance(parent, (ast.List, ast.Tuple)):
            to_delete = {
                change.node for change in changes if isinstance(change, Delete)
            }
            to_insert = {
                change.position: change.new_code
                for change in changes
                if isinstance(change, ListInsert)
            }

            def list_token_range(entry):
                r = list(source.asttokens().get_tokens(entry))
                return r[0], r[-1]

            generic_sequence_update(
                source,
                parent,
                brace_tokens(source, parent),
                [None if e in to_delete else list_token_range(e) for e in parent.elts],
                to_insert,
                recorder,
            )

        elif isinstance(parent, ast.Call):
            to_delete = {
                change.node for change in changes if isinstance(change, Delete)
            }
            atok = source.asttokens()

            def arg_token_range(node):
                if isinstance(node.parent, ast.keyword):
                    node = node.parent
                r = list(atok.get_tokens(node))
                return r[0], r[-1]

            braces_left = atok.next_token(list(atok.get_tokens(parent.func))[-1])
            assert braces_left.string == "("
            braces_right = list(atok.get_tokens(parent))[-1]
            assert braces_right.string == ")"

            to_insert = DefaultDict(list)

            for change in changes:
                if isinstance(change, CallArg):
                    if change.arg_name is not None:
                        position = (
                            change.arg_pos
                            if change.arg_pos is not None
                            else len(parent.args) + len(parent.keywords)
                        )
                        to_insert[position].append(
                            f"{change.arg_name} = {change.new_code}"
                        )
                    else:
                        assert change.arg_pos is not None
                        to_insert[change.arg_pos].append(change.new_code)

            generic_sequence_update(
                source,
                parent,
                (braces_left, braces_right),
                [
                    None if e in to_delete else arg_token_range(e)
                    for e in parent.args + [kw.value for kw in parent.keywords]
                ],
                to_insert,
                recorder,
            )

        elif isinstance(parent, ast.Dict):
            to_delete = {
                change.node for change in changes if isinstance(change, Delete)
            }
            to_insert = {
                change.position: [f"{key}: {value}" for key, value in change.new_code]
                for change in changes
                if isinstance(change, DictInsert)
            }

            def dict_token_range(key, value):
                return (
                    list(source.asttokens().get_tokens(key))[0],
                    list(source.asttokens().get_tokens(value))[-1],
                )

            generic_sequence_update(
                source,
                parent,
                brace_tokens(source, parent),
                [
                    None if value in to_delete else dict_token_range(key, value)
                    for key, value in zip(parent.keys, parent.values)
                ],
                to_insert,
                recorder,
            )

        else:
            assert False, parent



import ast
import copy
from typing import Any
from typing import Iterator

from .._adapter.adapter import AdapterContext
from .._adapter.adapter import get_adapter_type
from .._change import Change
from .._code_repr import code_repr
from .._exceptions import UsageError
from .._global_state import state
from .._sentinels import undefined
from .._types import SnapshotBase
from .._unmanaged import Unmanaged
from .._unmanaged import declare_unmanaged
from .._unmanaged import update_allowed


def clone(obj):
    new = copy.deepcopy(obj)
    if not obj == new:
        raise UsageError(
            f"""\
inline-snapshot uses `copy.deepcopy` to copy objects,
but the copied object is not equal to the original one:

value = {code_repr(obj)}
copied_value = copy.deepcopy(value)
assert value == copied_value

Please fix the way your object is copied or your __eq__ implementation.
"""
        )
    return new


def ignore_old_value():
    return state().update_flags.fix or state().update_flags.update


@declare_unmanaged
class GenericValue(SnapshotBase):
    _new_value: Any
    _old_value: Any
    _current_op = "undefined"
    _ast_node: ast.Expr
    _context: AdapterContext

    def _return(self, result, new_result=True):

        if not result:
            state().incorrect_values += 1
        flags = state().update_flags

        if flags.fix or flags.create or flags.update or self._old_value is undefined:
            return new_result
        return result

    @property
    def _file(self):
        return self._context.file

    def get_adapter(self, value):
        return get_adapter_type(value)(self._context)

    def _re_eval(self, value, context: AdapterContext):
        self._context = context

        def re_eval(old_value, node, value):
            if isinstance(old_value, Unmanaged):
                old_value.value = value
                return

            assert type(old_value) is type(value)

            adapter = self.get_adapter(old_value)
            if adapter is not None and hasattr(adapter, "items"):
                old_items = adapter.items(old_value, node)
                new_items = adapter.items(value, node)
                assert len(old_items) == len(new_items)

                for old_item, new_item in zip(old_items, new_items):
                    re_eval(old_item.value, old_item.node, new_item.value)

            else:
                if update_allowed(old_value):
                    if not old_value == value:
                        raise UsageError(
                            "snapshot value should not change. Use Is(...) for dynamic snapshot parts."
                        )
                else:
                    assert False, "old_value should be converted to Unmanaged"

        re_eval(self._old_value, self._ast_node, value)

    def _ignore_old(self):
        return (
            state().update_flags.fix
            or state().update_flags.update
            or state().update_flags.create
            or self._old_value is undefined
        )

    def _visible_value(self):
        if self._ignore_old():
            return self._new_value
        else:
            return self._old_value

    def _get_changes(self) -> Iterator[Change]:
        raise NotImplementedError()

    def _new_code(self):
        raise NotImplementedError()

    def __repr__(self):
        return repr(self._visible_value())

    def _type_error(self, op):
        __tracebackhide__ = True
        raise TypeError(
            f"This snapshot cannot be use with `{op}`, because it was previously used with `{self._current_op}`"
        )

    def __eq__(self, _other):
        __tracebackhide__ = True
        self._type_error("==")

    def __le__(self, _other):
        __tracebackhide__ = True
        self._type_error("<=")

    def __ge__(self, _other):
        __tracebackhide__ = True
        self._type_error(">=")

    def __contains__(self, _other):
        __tracebackhide__ = True
        self._type_error("in")

    def __getitem__(self, _item):
        __tracebackhide__ = True
        self._type_error("snapshot[key]")



import ast
from typing import Iterator

from .._change import Change
from .._change import Delete
from .._change import ListInsert
from .._change import Replace
from .._global_state import state
from .._sentinels import undefined
from .._utils import value_to_token
from .generic_value import GenericValue
from .generic_value import clone
from .generic_value import ignore_old_value


class CollectionValue(GenericValue):
    _current_op = "x in snapshot"

    def __contains__(self, item):
        if self._old_value is undefined:
            state().missing_values += 1

        if self._new_value is undefined:
            self._new_value = [clone(item)]
        else:
            if item not in self._new_value:
                self._new_value.append(clone(item))

        if ignore_old_value() or self._old_value is undefined:
            return True
        else:
            return self._return(item in self._old_value)

    def _new_code(self):
        return self._file._value_to_code(self._new_value)

    def _get_changes(self) -> Iterator[Change]:

        if self._ast_node is None:
            elements = [None] * len(self._old_value)
        else:
            assert isinstance(self._ast_node, ast.List)
            elements = self._ast_node.elts

        for old_value, old_node in zip(self._old_value, elements):
            if old_value not in self._new_value:
                yield Delete(
                    flag="trim",
                    file=self._file,
                    node=old_node,
                    old_value=old_value,
                )
                continue

            # check for update
            new_token = value_to_token(old_value)

            if (
                old_node is not None
                and self._file._token_of_node(old_node) != new_token
            ):
                new_code = self._file._token_to_code(new_token)

                yield Replace(
                    node=old_node,
                    file=self._file,
                    new_code=new_code,
                    flag="update",
                    old_value=old_value,
                    new_value=old_value,
                )

        new_values = [v for v in self._new_value if v not in self._old_value]
        if new_values:
            yield ListInsert(
                flag="fix",
                file=self._file,
                node=self._ast_node,
                position=len(self._old_value),
                new_code=[self._file._value_to_code(v) for v in new_values],
                new_values=new_values,
            )



from ._code_repr import HasRepr
from ._code_repr import customize_repr
from ._exceptions import UsageError
from ._external import external
from ._external import outsource
from ._inline_snapshot import snapshot
from ._is import Is
from ._types import Category
from ._types import Snapshot

__all__ = [
    "snapshot",
    "external",
    "outsource",
    "customize_repr",
    "HasRepr",
    "Is",
    "Category",
    "Snapshot",
    "UsageError",
]

__version__ = "0.24.0"



from __future__ import annotations

import ast
import typing
from dataclasses import dataclass

from inline_snapshot._source_file import SourceFile


def get_adapter_type(value):
    from inline_snapshot._adapter.generic_call_adapter import get_adapter_for_type

    adapter = get_adapter_for_type(type(value))
    if adapter is not None:
        return adapter

    if isinstance(value, list):
        from .sequence_adapter import ListAdapter

        return ListAdapter

    if type(value) is tuple:
        from .sequence_adapter import TupleAdapter

        return TupleAdapter

    if isinstance(value, dict):
        from .dict_adapter import DictAdapter

        return DictAdapter

    from .value_adapter import ValueAdapter

    return ValueAdapter


class Item(typing.NamedTuple):
    value: typing.Any
    node: ast.expr


@dataclass
class FrameContext:
    globals: dict
    locals: dict


@dataclass
class AdapterContext:
    file: SourceFile
    frame: FrameContext | None

    def eval(self, node):
        assert self.frame is not None

        return eval(
            compile(ast.Expression(node), self.file.filename, "eval"),
            self.frame.globals,
            self.frame.locals,
        )


class Adapter:
    context: AdapterContext

    def __init__(self, context: AdapterContext):
        self.context = context

    def get_adapter(self, old_value, new_value) -> Adapter:
        if type(old_value) is not type(new_value):
            from .value_adapter import ValueAdapter

            return ValueAdapter(self.context)

        adapter_type = get_adapter_type(old_value)
        if adapter_type is not None:
            return adapter_type(self.context)
        assert False

    def assign(self, old_value, old_node, new_value):
        raise NotImplementedError(self)

    def value_assign(self, old_value, old_node, new_value):
        from .value_adapter import ValueAdapter

        adapter = ValueAdapter(self.context)
        result = yield from adapter.assign(old_value, old_node, new_value)
        return result

    @classmethod
    def map(cls, value, map_function):
        raise NotImplementedError(cls)

    @classmethod
    def repr(cls, value):
        raise NotImplementedError(cls)


def adapter_map(value, map_function):
    return get_adapter_type(value).map(value, map_function)



import hashlib
import pathlib
import re
from typing import Optional
from typing import Set
from typing import Union

from . import _config
from ._global_state import state


class HashError(Exception):
    pass


class DiscStorage:
    def __init__(self, directory):
        self.directory = pathlib.Path(directory)

    def _ensure_directory(self):
        self.directory.mkdir(exist_ok=True, parents=True)
        gitignore = self.directory / ".gitignore"
        if not gitignore.exists():
            gitignore.write_text(
                "# ignore all snapshots which are not referred in the source\n*-new.*\n",
                "utf-8",
            )

    def save(self, name, data):
        assert "*" not in name
        self._ensure_directory()
        (self.directory / name).write_bytes(data)

    def read(self, name):
        return self._lookup_path(name).read_bytes()

    def prune_new_files(self):
        for file in self.directory.glob("*-new.*"):
            file.unlink()

    def list(self) -> Set[str]:
        if self.directory.exists():
            return {item.name for item in self.directory.iterdir()} - {".gitignore"}
        else:
            return set()

    def persist(self, name):
        try:
            file = self._lookup_path(name)
        except HashError:
            return
        if file.stem.endswith("-new"):
            stem = file.stem[:-4]
            file.rename(file.with_name(stem + file.suffix))

    def _lookup_path(self, name) -> pathlib.Path:
        files = list(self.directory.glob(name))

        if len(files) > 1:
            raise HashError(f"hash collision files={sorted(f.name for f in  files)}")

        if not files:
            raise HashError(f"hash {name!r} is not found in the DiscStorage")

        return files[0]

    def lookup_all(self, name) -> Set[str]:
        return {file.name for file in self.directory.glob(name)}

    def remove(self, name):
        self._lookup_path(name).unlink()


class external:
    def __init__(self, name: str):
        """External objects are used as a representation for outsourced data.
        You should not create them directly.

        The external data is by default stored inside `<pytest_config_dir>/.inline-snapshot/external`,
        where `<pytest_config_dir>` is replaced by the directory containing the Pytest configuration file, if any.
        To store data in a different location, set the `storage-dir` option in pyproject.toml.
        Data which is outsourced but not referenced in the source code jet has a '-new' suffix in the filename.

        Parameters:
            name: the name of the external stored object.
        """

        m = re.fullmatch(r"([0-9a-fA-F]*)\*?(\.[a-zA-Z0-9]*)", name)

        if m:
            self._hash, self._suffix = m.groups()
        else:
            raise ValueError(
                "path has to be of the form <hash>.<suffix> or <partial_hash>*.<suffix>"
            )

    @property
    def _path(self):
        return f"{self._hash}*{self._suffix}"

    def __repr__(self):
        """Returns the representation of the external object.

        The length of the hash can be specified in the
        [config](configuration.md).
        """
        hash = self._hash[: _config.config.hash_length]

        if len(hash) == 64:
            return f'external("{hash}{self._suffix}")'
        else:
            return f'external("{hash}*{self._suffix}")'

    def __eq__(self, other):
        """Two external objects are equal if they have the same hash and
        suffix."""
        if not isinstance(other, external):
            return NotImplemented

        min_hash_len = min(len(self._hash), len(other._hash))

        if self._hash[:min_hash_len] != other._hash[:min_hash_len]:
            return False

        if self._suffix != other._suffix:
            return False

        return True

    def _load_value(self):
        assert state().storage is not None
        return state().storage.read(self._path)


def outsource(data: Union[str, bytes], *, suffix: Optional[str] = None) -> external:
    """Outsource some data into an external file.

    ``` pycon
    >>> png_data = b"some_bytes"  # should be the replaced with your actual data
    >>> outsource(png_data, suffix=".png")
    external("212974ed1835*.png")

    ```

    Parameters:
        data: data which should be outsourced. strings are encoded with `"utf-8"`.

        suffix: overwrite file suffix. The default is `".bin"` if data is an instance of `#!python bytes` and `".txt"` for `#!python str`.

    Returns:
        The external data.
    """
    if isinstance(data, str):
        data = data.encode("utf-8")
        if suffix is None:
            suffix = ".txt"

    elif isinstance(data, bytes):
        if suffix is None:
            suffix = ".bin"
    else:
        raise TypeError("data has to be of type bytes | str")

    if not suffix or suffix[0] != ".":
        raise ValueError("suffix has to start with a '.' like '.png'")

    m = hashlib.sha256()
    m.update(data)
    hash = m.hexdigest()

    storage = state().storage

    assert storage is not None

    name = hash + suffix

    if not storage.lookup_all(name):
        path = hash + "-new" + suffix
        storage.save(path, data)

    return external(name)



from typing import Iterator
from typing import List

from inline_snapshot._adapter.adapter import Adapter

from .._change import Change
from .._compare_context import compare_only
from .._global_state import state
from .._sentinels import undefined
from .generic_value import GenericValue
from .generic_value import clone


class EqValue(GenericValue):
    _current_op = "x == snapshot"
    _changes: List[Change]

    def __eq__(self, other):
        if self._old_value is undefined:
            state().missing_values += 1

        if not compare_only() and self._new_value is undefined:
            self._changes = []
            adapter = Adapter(self._context).get_adapter(self._old_value, other)
            it = iter(adapter.assign(self._old_value, self._ast_node, clone(other)))
            while True:
                try:
                    self._changes.append(next(it))
                except StopIteration as ex:
                    self._new_value = ex.value
                    break

        return self._return(self._old_value == other, self._new_value == other)

    def _new_code(self):
        return self._file._value_to_code(self._new_value)

    def _get_changes(self) -> Iterator[Change]:
        return iter(self._changes)



from typing import Callable

from rich.console import Console

all_problems = set()


def raise_problem(message):
    all_problems.add(message)


def report_problems(console: Callable[[], Console]):

    global all_problems
    if not all_problems:
        return
    console().rule("[red]Problems")
    for problem in all_problems:
        console().print(f"{problem}")
        console().print()

    all_problems = set()



from __future__ import annotations

import ast
import warnings

from .._change import Delete
from .._change import DictInsert
from ..syntax_warnings import InlineSnapshotSyntaxWarning
from .adapter import Adapter
from .adapter import Item
from .adapter import adapter_map


class DictAdapter(Adapter):

    @classmethod
    def repr(cls, value):
        result = (
            "{"
            + ", ".join(f"{repr(k)}: {repr(value)}" for k, value in value.items())
            + "}"
        )

        if type(value) is not dict:
            result = f"{repr(type(value))}({result})"

        return result

    @classmethod
    def map(cls, value, map_function):
        return {k: adapter_map(v, map_function) for k, v in value.items()}

    @classmethod
    def items(cls, value, node):
        if node is None or not isinstance(node, ast.Dict):
            return [Item(value=value, node=None) for value in value.values()]

        result = []

        for value_key, node_key, node_value in zip(
            value.keys(), node.keys, node.values
        ):
            try:
                # this is just a sanity check, dicts should be ordered
                node_key = ast.literal_eval(node_key)
            except Exception:
                pass
            else:
                assert node_key == value_key

            result.append(Item(value=value[value_key], node=node_value))

        return result

    def assign(self, old_value, old_node, new_value):
        if old_node is not None:
            if not (
                isinstance(old_node, ast.Dict) and len(old_value) == len(old_node.keys)
            ):
                result = yield from self.value_assign(old_value, old_node, new_value)
                return result

            for key, value in zip(old_node.keys, old_node.values):
                if key is None:
                    warnings.warn_explicit(
                        "star-expressions are not supported inside snapshots",
                        filename=self.context.file._source.filename,
                        lineno=value.lineno,
                        category=InlineSnapshotSyntaxWarning,
                    )
                    return old_value

            for value, node in zip(old_value.keys(), old_node.keys):

                try:
                    # this is just a sanity check, dicts should be ordered
                    node_value = ast.literal_eval(node)
                except:
                    continue
                assert node_value == value

        result = {}
        for key, node in zip(
            old_value.keys(),
            (old_node.values if old_node is not None else [None] * len(old_value)),
        ):
            if key not in new_value:
                # delete entries
                yield Delete("fix", self.context.file._source, node, old_value[key])

        to_insert = []
        insert_pos = 0
        for key, new_value_element in new_value.items():
            if key not in old_value:
                # add new values
                to_insert.append((key, new_value_element))
                result[key] = new_value_element
            else:
                if isinstance(old_node, ast.Dict):
                    node = old_node.values[list(old_value.keys()).index(key)]
                else:
                    node = None
                # check values with same keys
                result[key] = yield from self.get_adapter(
                    old_value[key], new_value[key]
                ).assign(old_value[key], node, new_value[key])

                if to_insert:
                    new_code = [
                        (
                            self.context.file._value_to_code(k),
                            self.context.file._value_to_code(v),
                        )
                        for k, v in to_insert
                    ]
                    yield DictInsert(
                        "fix",
                        self.context.file._source,
                        old_node,
                        insert_pos,
                        new_code,
                        to_insert,
                    )
                    to_insert = []

                insert_pos += 1

        if to_insert:
            new_code = [
                (
                    self.context.file._value_to_code(k),
                    self.context.file._value_to_code(v),
                )
                for k, v in to_insert
            ]
            yield DictInsert(
                "fix",
                self.context.file._source,
                old_node,
                len(old_value),
                new_code,
                to_insert,
            )

        return result



import ast
import pathlib
from typing import Set

from executing import Source

from ._global_state import state
from ._rewrite_code import ChangeRecorder
from ._rewrite_code import end_of
from ._rewrite_code import start_of


def contains_import(tree, module, name):
    for node in tree.body:
        if (
            isinstance(node, ast.ImportFrom)
            and node.module == module
            and any(alias.name == name for alias in node.names)
        ):
            return True
    return False


def used_externals_in(source) -> Set[str]:
    tree = ast.parse(source)

    if not contains_import(tree, "inline_snapshot", "external"):
        return set()

    usages = []

    for node in ast.walk(tree):
        if (
            isinstance(node, ast.Call)
            and isinstance(node.func, ast.Name)
            and node.func.id == "external"
        ):
            usages.append(node)

    return {
        u.args[0].value
        for u in usages
        if u.args
        and isinstance(u.args[0], ast.Constant)
        and isinstance(u.args[0].value, str)
    }


def used_externals() -> Set[str]:
    result = set()
    for filename in state().files_with_snapshots:
        result |= used_externals_in(pathlib.Path(filename).read_text("utf-8"))

    return result


def unused_externals() -> Set[str]:
    storage = state().storage
    assert storage is not None
    unused_externals = storage.list()
    for name in used_externals():
        unused_externals -= storage.lookup_all(name)

    return unused_externals


def ensure_import(filename, imports, recorder: ChangeRecorder):
    source = Source.for_filename(filename)

    change = recorder.new_change()

    tree = source.tree
    token = source.asttokens()

    to_add = []

    for module, names in imports.items():
        for name in names:
            if not contains_import(tree, module, name):
                to_add.append((module, name))

    assert isinstance(tree, ast.Module)

    last_import = None
    for node in tree.body:
        if not isinstance(node, (ast.ImportFrom, ast.Import)):
            break
        last_import = node

    if last_import is None:
        position = start_of(tree.body[0].first_token)  # type: ignore
    else:
        last_token = last_import.last_token  # type: ignore
        while True:
            next_token = token.next_token(last_token)
            if last_token.end[0] == next_token.end[0]:
                last_token = next_token
            else:
                break
        position = end_of(last_token)

    code = ""
    for module, name in to_add:
        code += f"\nfrom {module} import {name}\n"

    change.insert(position, code, filename=filename)



import ast
import io
import token
import tokenize
from collections import namedtuple

from ._code_repr import code_repr


def normalize_strings(token_sequence):
    """Normalize string concattenanion.

    "a" "b" -> "ab"
    """

    current_string = None
    for t in token_sequence:
        if (
            t.type == token.STRING
            and not t.string.startswith(("'''", '"""', "b'''", 'b"""'))
            and t.string.startswith(("'", '"', "b'", 'b"'))
        ):
            if current_string is None:
                current_string = ast.literal_eval(t.string)
            else:
                current_string += ast.literal_eval(t.string)

            continue

        if current_string is not None:
            yield simple_token(token.STRING, repr(current_string))
            current_string = None

        yield t

    if current_string is not None:
        yield simple_token(token.STRING, repr(current_string))


def skip_trailing_comma(token_sequence):
    token_sequence = list(token_sequence)

    for index, token in enumerate(token_sequence):
        if index + 1 < len(token_sequence):
            next_token = token_sequence[index + 1]

            if token.string == "," and next_token.string in ("]", ")", "}"):
                continue
        yield token


def normalize(token_sequence):
    return skip_trailing_comma(normalize_strings(token_sequence))


ignore_tokens = (token.NEWLINE, token.ENDMARKER, token.NL)


# based on ast.unparse
def _str_literal_helper(string, *, quote_types):
    """Helper for writing string literals, minimizing escapes.

    Returns the tuple (string literal to write, possible quote types).
    """

    def escape_char(c):
        # \n and \t are non-printable, but we only escape them if
        # escape_special_whitespace is True
        if c in "\n\t":
            return c
        # Always escape backslashes and other non-printable characters
        if c == "\\" or not c.isprintable():
            return c.encode("unicode_escape").decode("ascii")
        if c == extra:
            return "\\" + c
        return c

    extra = ""
    if "'''" in string and '"""' in string:
        extra = '"' if string.count("'") >= string.count('"') else "'"

    escaped_string = "".join(map(escape_char, string))

    possible_quotes = [q for q in quote_types if q not in escaped_string]

    if escaped_string:
        # Sort so that we prefer '''"''' over """\""""
        possible_quotes.sort(key=lambda q: q[0] == escaped_string[-1])
        # If we're using triple quotes and we'd need to escape a final
        # quote, escape it
        if possible_quotes[0][0] == escaped_string[-1]:
            assert len(possible_quotes[0]) == 3
            escaped_string = escaped_string[:-1] + "\\" + escaped_string[-1]
    return escaped_string, possible_quotes


def triple_quote(string):
    """Write string literal value with a best effort attempt to avoid
    backslashes."""
    string, quote_types = _str_literal_helper(string, quote_types=['"""', "'''"])
    quote_type = quote_types[0]

    string = string.replace(" \n", " \\n\\\n")

    string = "\\\n" + string

    if not string.endswith("\n"):
        string = string + "\\\n"

    return f"{quote_type}{string}{quote_type}"


class simple_token(namedtuple("simple_token", "type,string")):

    def __eq__(self, other):
        if self.type == other.type == 3:
            if any(
                s.startswith(suffix)
                for s in (self.string, other.string)
                for suffix in ("f", "rf", "Rf", "F", "rF", "RF")
            ):
                return False

            return ast.literal_eval(self.string) == ast.literal_eval(
                other.string
            ) and self.string.replace("'", '"') == other.string.replace("'", '"')
        else:
            return super().__eq__(other)


def value_to_token(value):
    input = io.StringIO(code_repr(value))

    def map_string(tok):
        """Convert strings with newlines in triple quoted strings."""
        if tok.type == token.STRING:
            s = ast.literal_eval(tok.string)
            if isinstance(s, str) and (
                ("\n" in s and s[-1] != "\n") or s.count("\n") > 1
            ):
                # unparse creates a triple quoted string here,
                # because it thinks that the string should be a docstring
                triple_quoted_string = triple_quote(s)

                assert ast.literal_eval(triple_quoted_string) == s

                return simple_token(tok.type, triple_quoted_string)

        return simple_token(tok.type, tok.string)

    return [
        map_string(t)
        for t in tokenize.generate_tokens(input.readline)
        if t.type not in ignore_tokens
    ]



from __future__ import annotations

import ast
import warnings
from collections import defaultdict

from .._align import add_x
from .._align import align
from .._change import Delete
from .._change import ListInsert
from .._compare_context import compare_context
from ..syntax_warnings import InlineSnapshotSyntaxWarning
from .adapter import Adapter
from .adapter import Item
from .adapter import adapter_map


class SequenceAdapter(Adapter):
    node_type: type
    value_type: type
    braces: str
    trailing_comma: bool

    @classmethod
    def repr(cls, value):
        if len(value) == 1 and cls.trailing_comma:
            seq = repr(value[0]) + ","
        else:
            seq = ", ".join(map(repr, value))
        return cls.braces[0] + seq + cls.braces[1]

    @classmethod
    def map(cls, value, map_function):
        result = [adapter_map(v, map_function) for v in value]
        return cls.value_type(result)

    @classmethod
    def items(cls, value, node):
        if node is None or not isinstance(node, cls.node_type):
            return [Item(value=v, node=None) for v in value]

        assert len(value) == len(node.elts)

        return [Item(value=v, node=n) for v, n in zip(value, node.elts)]

    def assign(self, old_value, old_node, new_value):
        if old_node is not None:
            if not isinstance(
                old_node, ast.List if isinstance(old_value, list) else ast.Tuple
            ):
                result = yield from self.value_assign(old_value, old_node, new_value)
                return result

            for e in old_node.elts:
                if isinstance(e, ast.Starred):
                    warnings.warn_explicit(
                        "star-expressions are not supported inside snapshots",
                        filename=self.context.file.filename,
                        lineno=e.lineno,
                        category=InlineSnapshotSyntaxWarning,
                    )
                    return old_value

        with compare_context():
            diff = add_x(align(old_value, new_value))
        old = zip(
            old_value,
            old_node.elts if old_node is not None else [None] * len(old_value),
        )
        new = iter(new_value)
        old_position = 0
        to_insert = defaultdict(list)
        result = []
        for c in diff:
            if c in "mx":
                old_value_element, old_node_element = next(old)
                new_value_element = next(new)
                v = yield from self.get_adapter(
                    old_value_element, new_value_element
                ).assign(old_value_element, old_node_element, new_value_element)
                result.append(v)
                old_position += 1
            elif c == "i":
                new_value_element = next(new)
                new_code = self.context.file._value_to_code(new_value_element)
                result.append(new_value_element)
                to_insert[old_position].append((new_code, new_value_element))
            elif c == "d":
                old_value_element, old_node_element = next(old)
                yield Delete(
                    "fix",
                    self.context.file._source,
                    old_node_element,
                    old_value_element,
                )
                old_position += 1
            else:
                assert False

        for position, code_values in to_insert.items():
            yield ListInsert(
                "fix", self.context.file._source, old_node, position, *zip(*code_values)
            )

        return self.value_type(result)


class ListAdapter(SequenceAdapter):
    node_type = ast.List
    value_type = list
    braces = "[]"
    trailing_comma = False


class TupleAdapter(SequenceAdapter):
    node_type = ast.Tuple
    value_type = tuple
    braces = "()"
    trailing_comma = True



import tokenize
from pathlib import Path

from executing import Source

from inline_snapshot._format import enforce_formatting
from inline_snapshot._format import format_code
from inline_snapshot._utils import normalize
from inline_snapshot._utils import simple_token
from inline_snapshot._utils import value_to_token

from ._utils import ignore_tokens


class SourceFile:
    _source: Source

    def __init__(self, source: Source):
        self._source = source

    @property
    def filename(self):
        return self._source.filename

    def _format(self, text):
        if self._source is None or enforce_formatting():
            return text
        else:
            return format_code(text, Path(self._source.filename))

    def asttokens(self):
        return self._source.asttokens()

    def _token_to_code(self, tokens):
        return self._format(tokenize.untokenize(tokens)).strip()

    def _value_to_code(self, value):
        if value == "":
            # https://github.com/15r10nk/inline-snapshot/issues/258
            # this would otherwise cause a triple-quoted-string because black would format it as a docstring at the beginning of the code
            return '""'
        return self._token_to_code(value_to_token(value))

    def _token_of_node(self, node):

        return list(
            normalize(
                [
                    simple_token(t.type, t.string)
                    for t in self._source.asttokens().get_tokens(node)
                    if t.type not in ignore_tokens
                ]
            )
        )



import ast
from typing import Iterator

from .._adapter.adapter import AdapterContext
from .._change import Change
from .._change import Delete
from .._change import DictInsert
from .._global_state import state
from .._inline_snapshot import UndecidedValue
from .._sentinels import undefined
from .generic_value import GenericValue


class DictValue(GenericValue):
    _current_op = "snapshot[key]"

    def __getitem__(self, index):

        if self._new_value is undefined:
            self._new_value = {}

        if index not in self._new_value:
            old_value = self._old_value
            if old_value is undefined:
                state().missing_values += 1
                old_value = {}

            child_node = None
            if self._ast_node is not None:
                assert isinstance(self._ast_node, ast.Dict)
                if index in old_value:
                    pos = list(old_value.keys()).index(index)
                    child_node = self._ast_node.values[pos]

            self._new_value[index] = UndecidedValue(
                old_value.get(index, undefined), child_node, self._context
            )

        return self._new_value[index]

    def _re_eval(self, value, context: AdapterContext):
        super()._re_eval(value, context)

        if self._new_value is not undefined and self._old_value is not undefined:
            for key, s in self._new_value.items():
                if key in self._old_value:
                    s._re_eval(self._old_value[key], context)

    def _new_code(self):
        return (
            "{"
            + ", ".join(
                [
                    f"{self._file._value_to_code(k)}: {v._new_code()}"
                    for k, v in self._new_value.items()
                    if not isinstance(v, UndecidedValue)
                ]
            )
            + "}"
        )

    def _get_changes(self) -> Iterator[Change]:

        assert self._old_value is not undefined

        if self._ast_node is None:
            values = [None] * len(self._old_value)
        else:
            assert isinstance(self._ast_node, ast.Dict)
            values = self._ast_node.values

        for key, node in zip(self._old_value.keys(), values):
            if key in self._new_value:
                # check values with same keys
                yield from self._new_value[key]._get_changes()
            else:
                # delete entries
                yield Delete("trim", self._file, node, self._old_value[key])

        to_insert = []
        for key, new_value_element in self._new_value.items():
            if key not in self._old_value and not isinstance(
                new_value_element, UndecidedValue
            ):
                # add new values
                to_insert.append((key, new_value_element._new_code()))

        if to_insert:
            new_code = [(self._file._value_to_code(k), v) for k, v in to_insert]
            yield DictInsert(
                "create",
                self._file,
                self._ast_node,
                len(self._old_value),
                new_code,
                to_insert,
            )



from typing import Iterator

from .._change import Change
from .._change import Replace
from .._global_state import state
from .._sentinels import undefined
from .._utils import value_to_token
from .generic_value import GenericValue
from .generic_value import clone
from .generic_value import ignore_old_value


class MinMaxValue(GenericValue):
    """Generic implementation for <=, >="""

    @staticmethod
    def cmp(a, b):
        raise NotImplementedError

    def _generic_cmp(self, other):
        if self._old_value is undefined:
            state().missing_values += 1

        if self._new_value is undefined:
            self._new_value = clone(other)
            if self._old_value is undefined or ignore_old_value():
                return True
            return self._return(self.cmp(self._old_value, other))
        else:
            if not self.cmp(self._new_value, other):
                self._new_value = clone(other)

        return self._return(self.cmp(self._visible_value(), other))

    def _new_code(self):
        return self._file._value_to_code(self._new_value)

    def _get_changes(self) -> Iterator[Change]:
        new_token = value_to_token(self._new_value)
        if not self.cmp(self._old_value, self._new_value):
            flag = "fix"
        elif not self.cmp(self._new_value, self._old_value):
            flag = "trim"
        elif (
            self._ast_node is not None
            and self._file._token_of_node(self._ast_node) != new_token
        ):
            flag = "update"
        else:
            return

        new_code = self._file._token_to_code(new_token)

        yield Replace(
            node=self._ast_node,
            file=self._file,
            new_code=new_code,
            flag=flag,
            old_value=self._old_value,
            new_value=self._new_value,
        )


class MinValue(MinMaxValue):
    """
    handles:

    >>> snapshot(5) <= 6
    True

    >>> 6 >= snapshot(5)
    True

    """

    _current_op = "x >= snapshot"

    @staticmethod
    def cmp(a, b):
        return a <= b

    __le__ = MinMaxValue._generic_cmp


class MaxValue(MinMaxValue):
    """
    handles:

    >>> snapshot(5) >= 4
    True

    >>> 4 <= snapshot(5)
    True

    """

    _current_op = "x <= snapshot"

    @staticmethod
    def cmp(a, b):
        return a >= b

    __ge__ = MinMaxValue._generic_cmp



"""The following types are for type checking only."""

from typing import Literal
from typing import Protocol
from typing import TypeVar

T = TypeVar("T", covariant=True)


class SnapshotBase:
    pass


class Snapshot(Protocol[T]):
    """Can be used to annotate function arguments which accept snapshot
    values.

    You can annotate function arguments with `Snapshot[T]` to declare that a snapshot-value can be passed as function argument.
    `Snapshot[T]` is a type alias for `T`, which allows you to pass `int` values instead of `int` snapshots.


    Example:
    <!-- inline-snapshot: create fix trim first_block outcome-passed=2 -->
    ``` python
    from typing import Optional
    from inline_snapshot import snapshot, Snapshot

    # required snapshots


    def check_in_bounds(value, lower: Snapshot[int], upper: Snapshot[int]):
        assert lower <= value <= upper


    def test_numbers():
        for c in "hello world":
            check_in_bounds(ord(c), snapshot(32), snapshot(119))

        # use with normal values
        check_in_bounds(5, 0, 10)


    # optional snapshots


    def check_container(
        value,
        *,
        value_repr: Optional[Snapshot[str]] = None,
        length: Optional[Snapshot[int]] = None
    ):
        if value_repr is not None:
            assert repr(value) == value_repr

        if length is not None:
            assert len(value) == length


    def test_container():
        check_container([1, 2], value_repr=snapshot("[1, 2]"), length=snapshot(2))

        check_container({1, 1}, length=snapshot(1))
    ```
    """

    def __eq__(self, other: object, /) -> bool: ...  # pragma: no cover


Category = Literal["update", "fix", "create", "trim"]
"""See [categories](categories.md)"""



import ast
import os
import sys
from pathlib import Path

import pytest
from executing import is_pytest_compatible
from rich import box
from rich.console import Console
from rich.panel import Panel
from rich.prompt import Confirm
from rich.syntax import Syntax

from inline_snapshot.fix_pytest_diff import fix_pytest_diff

from . import _config
from . import _external
from . import _find_external
from ._change import apply_all
from ._code_repr import used_hasrepr
from ._find_external import ensure_import
from ._flags import Flags
from ._global_state import enter_snapshot_context
from ._global_state import leave_snapshot_context
from ._global_state import snapshot_env
from ._global_state import state
from ._inline_snapshot import used_externals
from ._problems import report_problems
from ._rewrite_code import ChangeRecorder
from ._snapshot.generic_value import GenericValue
from .pydantic_fix import pydantic_fix

pytest.register_assert_rewrite("inline_snapshot.extra")
pytest.register_assert_rewrite("inline_snapshot.testing._example")

if sys.version_info >= (3, 13):
    # fixes #186
    try:
        import readline  # noqa
    except (ImportError, ModuleNotFoundError):  # pragma: no cover
        # should fix #189 and #245
        pass


def pytest_addoption(parser, pluginmanager):
    group = parser.getgroup("inline-snapshot")

    group.addoption(
        "--inline-snapshot",
        metavar="(disable,short-report,report,review,create,update,trim,fix)*",
        dest="inline_snapshot",
        help="update specific snapshot values:\n"
        "disable: disable the snapshot logic\n"
        "short-report: show a short report\n"
        "report: show a longer report with a diff report\n"
        "review: allow to approve the changes interactive\n"
        "create: creates snapshots which are currently not defined\n"
        "update: update snapshots even if they are defined\n"
        "trim: changes the snapshot in a way which will make the snapshot more precise.\n"
        "fix: change snapshots which currently break your tests\n",
    )

    config_path = Path("pyproject.toml")
    if config_path.exists():
        config = _config.read_config(config_path)
        for name, value in config.shortcuts.items():
            value = ",".join(value)
            group.addoption(
                f"--{name}",
                action="store_const",
                const=value,
                dest="inline_snapshot",
                help=f'shortcut for "--inline-snapshot={value}"',
            )


categories = Flags.all().to_set()


def xdist_running(config):
    return (
        hasattr(config.option, "numprocesses")
        and config.option.numprocesses is not None
        and config.option.numprocesses != 0
    )


def is_ci_run():
    if bool(os.environ.get("PYCHARM_HOSTED", False)):
        # pycharm exports TEAMCITY_VERSION but luckily also PYCHARM_HOSTED,
        # which allows to detect the incorrect ci detection
        return False

    ci_env_vars = (
        "CI",
        "bamboo.buildKey",
        "BUILD_ID",
        "BUILD_NUMBER",
        "BUILDKITE",
        "CIRCLECI",
        "CONTINUOUS_INTEGRATION",
        "GITHUB_ACTIONS",
        "HUDSON_URL",
        "JENKINS_URL",
        "TEAMCITY_VERSION",
        "TRAVIS",
    )
    for var in ci_env_vars:
        if os.environ.get(var, False):
            return var
    return False


def is_implementation_supported():
    return sys.implementation.name == "cpython"


def pytest_configure(config):
    enter_snapshot_context()

    directory = config.rootpath
    while not (pyproject := directory / "pyproject.toml").exists():
        if directory == directory.parent:
            break
        directory = directory.parent
    _config.config = _config.Config()

    if is_pytest_compatible():
        _config.config.default_flags_tui = ["create", "review"]
        _config.config.default_flags = ["report"]

    _config.read_config(pyproject, _config.config)

    console = Console()
    if is_ci_run():
        default_flags = {"disable"}
    elif console.is_terminal:
        default_flags = _config.config.default_flags_tui
    else:
        default_flags = _config.config.default_flags

    env_var = "INLINE_SNAPSHOT_DEFAULT_FLAGS"
    if env_var in os.environ:
        default_flags = os.environ[env_var].split(",")

    if config.option.inline_snapshot is None:
        flags = set(default_flags)
    else:
        flags = config.option.inline_snapshot.split(",")
        flags = {flag for flag in flags if flag}
        if xdist_running(config) and flags - {"disable"}:
            raise pytest.UsageError(
                f"--inline-snapshot={','.join(flags)} can not be combined with xdist"
            )
    state().flags = flags

    unknown_flags = flags - categories - {"disable", "review", "report", "short-report"}
    if unknown_flags:
        raise pytest.UsageError(
            f"--inline-snapshot={','.join(sorted(unknown_flags))} is a unknown flag"
        )

    if "disable" in flags and flags != {"disable"}:
        raise pytest.UsageError(
            f"--inline-snapshot=disable can not be combined with other flags ({', '.join(flags-{'disable'})})"
        )

    if xdist_running(config) or not is_implementation_supported():
        state().active = False

    elif flags & {"review"}:
        state().active = True

        state().update_flags = Flags.all()
    else:

        state().active = "disable" not in flags

        state().update_flags = Flags(flags & categories)

    external_storage = (
        _config.config.storage_dir or config.rootpath / ".inline-snapshot"
    ) / "external"

    state().storage = _external.DiscStorage(external_storage)

    if flags - {"short-report", "disable"} and not is_pytest_compatible():

        # hack to disable the assertion rewriting
        # I found no other way because the hook gets installed early
        sys.meta_path = [
            e for e in sys.meta_path if type(e).__name__ != "AssertionRewritingHook"
        ]

    pydantic_fix()

    fix_pytest_diff()

    state().storage.prune_new_files()


def is_xfail(request):
    if not "xfail" in request.keywords:
        return False
    xfail = request.keywords["xfail"]
    if xfail.args and xfail.args[0] == False:
        return False
    return True


@pytest.fixture(autouse=True)
def snapshot_check(request):
    state().missing_values = 0
    state().incorrect_values = 0

    if is_xfail(request):
        with snapshot_env() as local_state:
            local_state.active = False
            yield
        return
    else:
        yield

    missing_values = state().missing_values
    incorrect_values = state().incorrect_values

    extra = "\nIf you just created this value with --snapshot=create, the value is now created and you can ignore this message."

    if missing_values != 0:
        pytest.fail(
            (
                "your snapshot is missing one value."
                if missing_values == 1
                else f"your snapshot is missing {missing_values} values."
            )
            + extra,
            pytrace=False,
        )

    if incorrect_values != 0:
        pytest.fail(
            "some snapshots in this test have incorrect values." + extra,
            pytrace=False,
        )


def pytest_assertrepr_compare(config, op, left, right):

    results = []
    if isinstance(left, GenericValue):
        results = config.hook.pytest_assertrepr_compare(
            config=config, op=op, left=left._visible_value(), right=right
        )

    if isinstance(right, GenericValue):
        results = config.hook.pytest_assertrepr_compare(
            config=config, op=op, left=left, right=right._visible_value()
        )

    external_used = False
    if isinstance(right, _external.external):
        external_used = True
        if right._suffix == ".txt":
            right = right._load_value().decode()
        else:
            right = right._load_value()

    if isinstance(left, _external.external):
        external_used = True
        if left._suffix == ".txt":
            left = left._load_value().decode()
        else:
            left = left._load_value()

    if external_used:
        results = config.hook.pytest_assertrepr_compare(
            config=config, op=op, left=left, right=right
        )

    if results:
        return results[0]


def call_once(f):
    called = False
    result = None

    def w():
        nonlocal called
        nonlocal result
        if not called:
            result = f()
            called = True
        return result

    return w


def link(text, link=None):
    return f"[italic blue link={link or text}]{text}[/]"


def category_link(category):
    return link(
        category,
        f"https://15r10nk.github.io/inline-snapshot/latest/categories/#{category}",
    )


@pytest.hookimpl(trylast=True)
def pytest_sessionfinish(session, exitstatus):
    try:
        config = session.config

        @call_once
        def console():
            con = Console(highlight=False)
            con.print("\n")
            con.rule("[blue]inline-snapshot", characters="═")

            return con

        disable_info = "This means that tests with snapshots will continue to run, but snapshot(x) will only return x and inline-snapshot will not be able to fix snapshots or generate reports."

        if xdist_running(config):
            if state().flags != {"disable"}:
                console().print(
                    f"INFO: inline-snapshot was disabled because you used xdist. {disable_info}\n"
                )
            return

        if env_var := is_ci_run():
            if state().flags == {"disable"}:
                console().print(
                    f'INFO: CI run was detected because environment variable "{env_var}" was defined. '
                    f"inline-snapshot runs with --inline-snapshot=disable by default in CI. {disable_info} You can change this by using --inline-snasphot=report for example.\n"
                )
                return

        if not is_implementation_supported():
            if state().flags != {"disable"}:
                console().print(
                    f"INFO: inline-snapshot was disabled because {sys.implementation.name} is not supported. {disable_info}\n"
                )
            return

        if not state().active:
            return

        capture = config.pluginmanager.getplugin("capturemanager")

        # --inline-snapshot

        def apply_changes(flag):
            if flag in state().flags:
                console().print(
                    f"These changes will be applied, because you used {category_link(flag)}",
                    highlight=False,
                )
                console().print()
                return True
            if "review" in state().flags:
                result = Confirm.ask(
                    f"Do you want to {category_link(flag)} these snapshots?",
                    default=False,
                )
                console().print()
                return result
            else:
                console().print("These changes are not applied.")
                console().print(
                    f"Use [bold]--inline-snapshot={category_link(flag)} to apply them, "
                    "or use the interactive mode with [b]--inline-snapshot=[italic blue link=https://15r10nk.github.io/inline-snapshot/latest/pytest/#-inline-snapshotreview]review[/][/]",
                    highlight=False,
                )
                console().print()
                return False

        # auto mode
        changes = {f: [] for f in Flags.all()}

        snapshot_changes = {f: 0 for f in Flags.all()}

        for snapshot in state().snapshots.values():
            all_categories = set()
            for change in snapshot._changes():
                changes[change.flag].append(change)
                all_categories.add(change.flag)

            for category in all_categories:
                snapshot_changes[category] += 1

        capture.suspend_global_capture(in_=True)
        try:

            if "short-report" in state().flags:

                def report(flag, message, message_n):
                    num = snapshot_changes[flag]

                    if num and not getattr(state().update_flags, flag):
                        console().print(
                            message if num == 1 else message.format(num=num),
                            highlight=False,
                        )

                report(
                    "fix",
                    "Error: one snapshot has incorrect values ([b]--inline-snapshot=fix[/])",
                    "Error: {num} snapshots have incorrect values ([b]--inline-snapshot=fix[/])",
                )

                report(
                    "trim",
                    "Info: one snapshot can be trimmed ([b]--inline-snapshot=trim[/])",
                    "Info: {num} snapshots can be trimmed ([b]--inline-snapshot=trim[/])",
                )

                report(
                    "create",
                    "Error: one snapshot is missing a value ([b]--inline-snapshot=create[/])",
                    "Error: {num} snapshots are missing values ([b]--inline-snapshot=create[/])",
                )

                report(
                    "update",
                    "Info: one snapshot changed its representation ([b]--inline-snapshot=update[/])",
                    "Info: {num} snapshots changed their representation ([b]--inline-snapshot=update[/])",
                )

                if sum(snapshot_changes.values()) != 0:
                    console().print(
                        "\nYou can also use [b]--inline-snapshot=review[/] to approve the changes interactively",
                        highlight=False,
                    )

                return

            if not is_pytest_compatible():
                assert not any(
                    type(e).__name__ == "AssertionRewritingHook" for e in sys.meta_path
                )

            used_changes = []
            for flag in Flags.all():
                if not changes[flag]:
                    continue

                if not {"review", "report", flag} & state().flags:
                    continue

                @call_once
                def header():
                    console().rule(f"[yellow bold]{flag.capitalize()} snapshots")

                if (
                    flag == "update"
                    and not _config.config.show_updates
                    and not "update" in state().flags
                ):
                    continue

                cr = ChangeRecorder()
                apply_all(used_changes, cr)
                cr.virtual_write()
                apply_all(changes[flag], cr)

                any_changes = False

                for file in cr.files():
                    diff = file.diff()
                    if diff:
                        header()
                        name = file.filename.relative_to(Path.cwd())
                        console().print(
                            Panel(
                                Syntax(diff, "diff", theme="ansi_light"),
                                title=str(name),
                                box=(
                                    box.ASCII
                                    if os.environ.get("TERM", "") == "unknown"
                                    else box.ROUNDED
                                ),
                            )
                        )
                        any_changes = True

                if any_changes and apply_changes(flag):
                    used_changes += changes[flag]

            report_problems(console)

            if used_changes:
                cr = ChangeRecorder()
                apply_all(used_changes, cr)

                for test_file in cr.files():
                    tree = ast.parse(test_file.new_code())
                    used = used_externals(tree)

                    required_imports = []

                    if used:
                        required_imports.append("external")

                    if used_hasrepr(tree):
                        required_imports.append("HasRepr")

                    if required_imports:
                        ensure_import(
                            test_file.filename,
                            {"inline_snapshot": required_imports},
                            cr,
                        )

                    for external_name in used:
                        state().storage.persist(external_name)

                cr.fix_all()

            unused_externals = _find_external.unused_externals()

            if unused_externals and state().update_flags.trim:
                for name in unused_externals:
                    assert state().storage
                    state().storage.remove(name)
                console().print(f"removed {len(unused_externals)} unused externals\n")
        finally:
            capture.resume_global_capture()

        return
    finally:
        leave_snapshot_context()



try:
    import dirty_equals  # type: ignore
except ImportError:  # pragma: no cover

    def is_dirty_equal(value):
        return False

else:

    def is_dirty_equal(value):
        return isinstance(value, dirty_equals.DirtyEquals) or (
            isinstance(value, type) and issubclass(value, dirty_equals.DirtyEquals)
        )


def update_allowed(value):
    global unmanaged_types
    return not (is_dirty_equal(value) or isinstance(value, tuple(unmanaged_types)))  # type: ignore


unmanaged_types = []


def is_unmanaged(value):
    return not update_allowed(value)


def declare_unmanaged(data_type):
    global unmanaged_types
    unmanaged_types.append(data_type)
    return data_type


class Unmanaged:
    def __init__(self, value):
        self.value = value

    def __eq__(self, other):
        assert not isinstance(other, Unmanaged)

        return self.value == other

    def __repr__(self):
        return repr(self.value)


def map_unmanaged(value):
    if is_unmanaged(value):
        return Unmanaged(value)
    else:
        return value
